@@ -9,6 +9,7 @@ import xml.sax as sax
 import os
 import io
 import shutil
+import dataclasses
 
 from firewall import config
 from firewall.functions import (
@@ -44,27 +45,27 @@ class Zone(IO_Object):
     priority_max = 32767
     priority_default = DEFAULT_ZONE_PRIORITY
 
-    IMPORT_EXPORT_STRUCTURE = (
-        ("version", ""),  # s
-        ("short", ""),  # s
-        ("description", ""),  # s
-        ("UNUSED", False),  # b
-        ("target", ""),  # s
-        ("services", [""]),  # as
-        ("ports", [("", "")]),  # a(ss)
-        ("icmp_blocks", [""]),  # as
-        ("masquerade", False),  # b
-        ("forward_ports", [("", "", "", "")]),  # a(ssss)
-        ("interfaces", [""]),  # as
-        ("sources", [""]),  # as
-        ("rules_str", [""]),  # as
-        ("protocols", [""]),  # as
-        ("source_ports", [("", "")]),  # a(ss)
-        ("icmp_block_inversion", False),  # b
-        ("forward", True),  # b
-        ("ingress_priority", 0),  # i
-        ("egress_priority", 0),  # i
-    )
+    IMPORT_EXPORT_STRUCTURE = {
+        "version": "",  # s
+        "short": "",  # s
+        "description": "",  # s
+        "UNUSED": False,  # b
+        "target": "",  # s
+        "services": [""],  # as
+        "ports": [("", "")],  # a(ss)
+        "icmp_blocks": [""],  # as
+        "masquerade": False,  # b
+        "forward_ports": [("", "", "", "")],  # a(ssss)
+        "interfaces": [""],  # as
+        "sources": [""],  # as
+        "rules_str": [""],  # as
+        "protocols": [""],  # as
+        "source_ports": [("", "")],  # a(ss)
+        "icmp_block_inversion": False,  # b
+        "forward": True,  # b
+        "ingress_priority": 0,  # i
+        "egress_priority": 0,  # i
+    }
     ADDITIONAL_ALNUM_CHARS = ["_", "-", "/"]
     PARSER_REQUIRED_ELEMENT_ATTRS = {
         "short": None,
@@ -109,12 +110,13 @@ class Zone(IO_Object):
         "log": ["prefix", "level"],
         "nflog": ["group", "prefix", "queue-size"],
         "reject": ["type"],
+        "limit": ["burst"],
         "tcp-mss-clamp": ["value"],
     }
 
     @staticmethod
     def index_of(element):
-        for i, (el, dummy) in enumerate(Zone.IMPORT_EXPORT_STRUCTURE):
+        for i, el in enumerate(Zone.IMPORT_EXPORT_STRUCTURE):
             if el == element:
                 return i
         raise FirewallError(errors.UNKNOWN_ERROR, "index_of()")
@@ -136,8 +138,7 @@ class Zone(IO_Object):
         self.source_ports = []
         self.interfaces = []
         self.sources = []
-        self.rules = []
-        self.rules_str = []
+        self.rules = set()
         self.icmp_block_inversion = False
         self.combined = False
         self.applied = False
@@ -160,19 +161,22 @@ class Zone(IO_Object):
         del self.source_ports[:]
         del self.interfaces[:]
         del self.sources[:]
-        del self.rules[:]
-        del self.rules_str[:]
+        self.rules.clear()
         self.icmp_block_inversion = False
         self.combined = False
         self.applied = False
         self.ingress_priority = self.priority_default
         self.egress_priority = self.priority_default
 
+    def __getattr__(self, name):
+        if name == "rules_str":
+            return [str(r) for r in sorted(self.rules)]
+        else:
+            return getattr(super(Zone, self), name)
+
     def __setattr__(self, name, value):
         if name == "rules_str":
-            self.rules = [rich.Rich_Rule(rule_str=s) for s in value]
-            # must convert back to string to get the canonical string.
-            super(Zone, self).__setattr__(name, [str(s) for s in self.rules])
+            self.rules = set([rich.Rich_Rule(rule_str=s) for s in value])
         else:
             super(Zone, self).__setattr__(name, value)
 
@@ -309,8 +313,7 @@ class Zone(IO_Object):
             if port not in self.source_ports:
                 self.source_ports.append(port)
         for rule in zone.rules:
-            self.rules.append(rule)
-            self.rules_str.append(str(rule))
+            self.rules.add(rule)
         if zone.icmp_block_inversion:
             self.icmp_block_inversion = True
 
@@ -322,13 +325,10 @@ class zone_ContentHandler(IO_Object_ContentHandler):
     def __init__(self, item):
         IO_Object_ContentHandler.__init__(self, item)
         self._rule = None
-        self._rule_error = False
         self._limit_ok = None
 
     def startElement(self, name, attrs):
         IO_Object_ContentHandler.startElement(self, name, attrs)
-        if self._rule_error:
-            return
 
         self.item.parser_check_element_attrs(name, attrs)
 
@@ -356,35 +356,24 @@ class zone_ContentHandler(IO_Object_ContentHandler):
                 self.item.egress_priority = int(attrs["egress-priority"])
 
         elif name == "forward":
-            if self.item.forward:
-                log.warning("Forward already set, ignoring.")
-            else:
-                self.item.forward = True
+            self.item.forward = True
 
         elif name == "interface":
             if self._rule:
-                log.warning("Invalid rule: interface use in rule.")
-                self._rule_error = True
-                return
-            # zone bound to interface
-            if "name" not in attrs:
-                log.warning("Invalid interface: Name missing.")
-                self._rule_error = True
-                return
+                raise FirewallError(
+                    errors.INVALID_RULE,
+                    f"Interface is not valid in rule '{str(self._rule)}'.",
+                )
             if attrs["name"] not in self.item.interfaces:
                 self.item.interfaces.append(attrs["name"])
-            else:
-                log.warning("Interface '%s' already set, ignoring.", attrs["name"])
 
         elif name == "source":
             if self._rule:
                 if self._rule.source:
-                    log.warning(
-                        "Invalid rule: More than one source in rule '%s', ignoring.",
-                        str(self._rule),
+                    raise FirewallError(
+                        errors.INVALID_RULE,
+                        f"More than one source in rule '{str(self._rule)}'.",
                     )
-                    self._rule_error = True
-                    return
                 invert = False
                 if "invert" in attrs and attrs["invert"].lower() in ["yes", "true"]:
                     invert = True
@@ -395,21 +384,25 @@ class zone_ContentHandler(IO_Object_ContentHandler):
                     mac = attrs["mac"]
                 if "ipset" in attrs:
                     ipset = attrs["ipset"]
-                self._rule.source = rich.Rich_Source(addr, mac, ipset, invert=invert)
+                self._rule = dataclasses.replace(
+                    self._rule, source=rich.Rich_Source(addr, mac, ipset, invert=invert)
+                )
                 return
             # zone bound to source
             if "address" not in attrs and "ipset" not in attrs:
-                log.warning("Invalid source: No address no ipset.")
-                return
+                raise FirewallError(
+                    errors.INVALID_SOURCE, "No address or ipset specified."
+                )
             if "address" in attrs and "ipset" in attrs:
-                log.warning("Invalid source: Address and ipset.")
-                return
+                raise FirewallError(
+                    errors.INVALID_SOURCE, "Both address and ipset (can only use one)."
+                )
             if "family" in attrs:
                 log.warning(
                     "Ignoring deprecated attribute family='%s'", attrs["family"]
                 )
             if "invert" in attrs:
-                log.warning("Invalid source: Invertion not allowed here.")
+                raise FirewallError(errors.INVALID_SOURCE, "Invert not allowed.")
                 return
             if "address" in attrs:
                 if (
@@ -422,23 +415,16 @@ class zone_ContentHandler(IO_Object_ContentHandler):
                 entry = "ipset:%s" % attrs["ipset"]
                 if entry not in self.item.sources:
                     self.item.sources.append(entry)
-                else:
-                    log.warning("Source '%s' already set, ignoring.", attrs["address"])
             if "address" in attrs:
                 entry = attrs["address"]
                 if entry not in self.item.sources:
                     self.item.sources.append(entry)
-                else:
-                    log.warning("Source '%s' already set, ignoring.", attrs["address"])
 
         elif name == "icmp-block-inversion":
-            if self.item.icmp_block_inversion:
-                log.warning("Icmp-Block-Inversion already set, ignoring.")
-            else:
-                self.item.icmp_block_inversion = True
+            self.item.icmp_block_inversion = True
 
         else:
-            log.warning("Unknown XML element '%s'", name)
+            raise FirewallError(errors.INVALID_ZONE, f"Unknown XML element '{name}'.")
             return
 
     def endElement(self, name):
